#!/usr/bin/python2

"""This is the Client.  It controls the flow of the program."""

import sys
import socket
from StringIO import StringIO

from AutomaticClass import AutomaticClass
import DataModel
from DataModel import game
from Strategy import Strategy

class Client(AutomaticClass):

    """This is the Client.  It controls the flow of the program.
    
    The following methods are defined:

    run -- Connect to the server, and play the game.
    
    The following private methods are defined:

    _receiveWorld -- Receive initialization about the World.

    _receiveRobotSelf -- Receive initialization about the robot this game 
    controls.

    _receiveUpdate -- Receive an update on all the Robots.

    _receivePackages -- Receive a list of the packages at the current location.

    _sendCommand -- Send a command to the server.

    _readLine -- Return a line of input from the server, sans newline.

    The following private variables are used:

    _robotSelfDict -- _receiveRobotSelf doesn't provide enough information to 
    construct a RobotSelf.  Hence, I'll have to store some data here until
    I can finish the job in _receiveUpdate.

    """

    def __init__(self, **kargs):
        fields = {
            "host": None,
            "port": None,
            "_socket": None
        } 
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 

    def run(self):
        """Connect to the server, and play the game."""
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.host, self.port))
        self._socket.send("Player\n")
        try: 
            self._receiveWorld()
            self._receiveRobotSelf()
            self._receiveUpdate()
            while 1:
                self._receivePackages()
                self._sendCommand()
                self._receiveUpdate()
        except EOFError:
            pass
        self._socket.close()

    def _receiveWorld(self):
        """Receive initialization about the World.
        
        Remember, the World is 1 based, but Python, in general, is 0 based.
        
        """
        (width, height) = map(int, self._readLine().split(" "))
        world = DataModel.World(width = width, height = height)
        for y in range(0, height):
            line = self._readLine()
            assert len(line) == width
            for x in range(0, width):
                world[y + 1][x + 1] = DataModel.Square(terrain = line[x])
        game.world = world

    def _receiveRobotSelf(self):
        """Receive initialization about the robot this game controls."""
        (id, maximumCarryingCapacity, money) = map(
            int, self._readLine().split(" "))
        self._robotSelfDict = {
            "id": id, 
            "maximumCarryingCapacity": maximumCarryingCapacity, 
            "money": money
        }

    def _receiveUpdate(self):

        """Receive an update on all the Robots."""

        line = self._readLine()
        assert line != ""
        pieces = line.split(" ")
        updatedRobots = []
        while pieces:
            token, pieces = pieces[0], pieces[1:] 
            firstLetter = token[0]

            # Receive the id of a robot.

            if firstLetter == "#":
                id = int(token[1:])
                updatedRobots.append(id)
                if game.robots.has_key(id):
                    robot = game.robots[id]
                else:

                    # Hmm, I've never heard of this guy.  He must be new.
                    # He may even be the rest of the data concerning me! ;)

                    assert pieces[0] == "X" and pieces[2] == "Y"
                    x, y, pieces = int(pieces[1]), int(pieces[3]), pieces[4:]
                    if id == self._robotSelfDict["id"]:
                        d = self._robotSelfDict
                        robot = DataModel.RobotSelf(
                            id = id, location = (x, y),
                            maximumCarryingCapacity = 
                                d["maximumCarryingCapacity"], 
                            money = d["money"])
                    else:
                        robot = DataModel.RobotOther(
                            id = id, location = (x, y))

            # The robot has moved.

            elif firstLetter in DataModel.DIRECTIONS:
                robot.move(firstLetter)

            # Handle picking and dropping.

            elif firstLetter in ["P", "D"]:
                id, pieces = pieces[0], pieces[1:] 
                if game.packages.has_key(id):
                    package = game.packages[id]
                else:
                    package = DataModel.Package(id = id)
                if firstLetter == "P": 
                    robot.pick(package)
                else: 
                    robot.drop(package)

        # Any robot not updated is dead.  Delete it.

        map(lambda i: i.destruct(), 
            [i for i in game.robots if not i in updatedRobots])

        print game.world
       
    def _receivePackages(self):
        """Receive a list of the packages at the current location."""
        square = game.robotSelf.square
        square.packagesEncountered = []
        line = self._readLine()
        if not line: 
            return # Otherwise, line.split will give you [''].
        pieces = line.split(" ")
        while pieces:
            (id, x, y, weight), pieces = pieces[:4], pieces[4:]
            package = DataModel.Package(id = id, weight = weight, 
                destination = (x, y))
            square.packagesEncountered.append(package)

    def _sendCommand(self):
        """Send a command to the server."""
        self._socket.send("1 Move %s\n" % Strategy().getMove())

    def _readLine(self):
        """Return a line of input from the server, sans newline.
        
        Upon EOF, raise EOFError.
        
        """
        buf = StringIO()
        while 1:
            recv = self._socket.recv(1)
            if len(recv) == 0:
                raise EOFError
            buf.write(recv)
            value = buf.getvalue()
            if value[-1] == "\n":
                return value[:-1]

# Do some testing.
if __name__ == '__main__':
    Client(host = sys.argv[1], port = int(sys.argv[2])).run()
