"""This file contains the data model for the game.

The following constants are defined:

OPEN_SPACE, WALL, WATER, HOME_BASE -- These are the possible values for the 
terrain on a Square.  I'll use strings to define them so that they are easier 
to read and write.

NORTH, EAST, SOUTH, WEST -- These are the directions in which the robot can 
move.  Again, I'll use strings to define them, as above.

DIRECTIONS -- A list containing above directions.

The following module level variables are defined:

game -- This is an instance of the Game class.

The following classes are defined: 

Square -- This encapsulates all the things that can be on a part of the World.

World -- This behaves like a two-dimensional array.

RobotBase -- This is the base class for robots.

RobotSelf -- The robot that this game controls.

RobotOther -- A robot that this game competes against.

Package -- This is a simple class to represent a package.

Game -- This class ties a lot of the other classes together.

The following methods are defined:

move -- Given a location and a direction, return the new location.

inverseMove -- Given two locations, return a direction.

getAdjacent -- Return the locations adjecent to a given location.

"""

from AutomaticClass import AutomaticClass
from UserList import UserList
import operator

OPEN_SPACE = "."
WALL = "#"
WATER = "~"
HOME_BASE = "@"

NORTH = "N"
EAST = "E"
SOUTH = "S"
WEST = "W"
DIRECTIONS = [NORTH, EAST, SOUTH, WEST]

def move(location, direction):
    """Given a location and a direction, return the new location.
    
    This is not constrained.
    
    """
    delta = {
        NORTH: (0, 1),
        EAST: (1, 0),
        SOUTH: (0, -1),
        WEST: (-1, 0)
    }
    deltaTuple = delta[direction]
    return (location[0] + deltaTuple[0], 
            location[1] + deltaTuple[1])

def inverseMove(start, finish):
    """Given two locations, return a direction."""
    if start[0] < finish[0]:
        return EAST
    if start[0] > finish[0]:
        return WEST
    if start[1] < finish[1]:
        return NORTH
    return SOUTH

def getAdjacent(location):
    """Return the locations adjecent to a given location."""
    return [move(location, i) for i in DIRECTIONS]

class Square(AutomaticClass):

    """This encapsulates all the things that can be on a part of the World.
    
    __str__ -- Return a user-friendly string representation of this class.
    
    """

    def __init__(self, **kargs):
        fields = {
            "terrain": None,
            "robot": None,
            # Unless we are actually on this Square, packagesEncountered will 
            # contain stale, incomplete information.
            "packagesEncountered": [],
            # The following are used by the Strategy module.  They will be
            # updated each time we need to make a move.
            "shortestPath": None,
            "weightedValue": None
        } 
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 

    def __str__(self):
        """Return a user-friendly string representation of this class."""
        if self.robot:
            robot = "R"
        else:
            robot = " "
        if len(self.packagesEncountered):
            packagesEncountered = "P"
        else:
            packagesEncountered = " "
        return "%s%s%s" % (self.terrain, robot, packagesEncountered)

class World(UserList, AutomaticClass):
    
    """This behaves like a two-dimensional array.
    
    When you look up a Square, use the notation world[y][x].

    The following methods are defined:

    __init__ -- Create a list of lists of the correct size.
    
    __str__ -- Return a user-friendly string representation of this class.

    at -- Given a location, (x, y), return world[y][x].

    """

    __repr__ = AutomaticClass.__repr__ # Override UserList.__repr__.

    def __init__(self, **kargs):

        """Create a list of lists of the correct size.

        Each Square will be initialized to have terrain set to OPEN_SPACE.
        Because the coordinate system of the game is 1 based, and because the
        world is effectively surrounded by walls, I will literally include the
        walls (hence the difference between width and actualWidth and the 2's).
        This effectively shifts my model for the world so that it is 1 based as
        well, instead of 0 based (which Python lists are, by default).

        """

        fields = {
            "width": None,        # Don't alter.
            "height": None,       # Don't alter.
            "actualWidth": None,  # Don't alter.  Don't initialize.
            "actualHeight": None  # Don't alter.  Don't initialize.
        }
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 

        self.actualWidth = self.width + 2
        self.actualHeight = self.height + 2
        rows = []
        for y in range(0, self.actualHeight):
            cols = []
            for x in range(0, self.actualWidth):
                cols.append(Square(terrain = OPEN_SPACE))
            rows.append(cols)
        UserList.__init__(self, rows)
        wall = Square(terrain = WALL) # Reuse it.
        for x in range(0, self.actualWidth): 
            self[0][x] = wall
            self[self.actualHeight - 1][x] = wall
        for y in range(0, self.actualHeight):
            self[y][0] = wall
            self[y][self.actualWidth - 1] = wall

    def __str__(self):
        """Return a user-friendly string representation of this class."""
        s = ""
        for y in range(0, self.actualHeight):
            y = self.actualHeight - 1 - y # Invert.
            for x in range(0, self.actualWidth):
                s += "%s" % self[y][x]
            s += "\n"
        return s[:-1]

    def at(self, location):
        """Given a location, (x, y), return world[y][x]."""
        return self[location[1]][location[0]]

class RobotBase(AutomaticClass):

    """This is the base class for robots.

    The following methods are defined:

    __init__ -- Put self into game.robots and into the World.

    __getattr__ -- square is an attribute that returns the square we're on.

    destruct -- Remove this robot from game.robots and the World.  Then del it.

    move -- Move the robot to an adjacent square.

    pick -- Add a package to our list of packages.

    drop -- Drop a package from our list of packages.

    """

    def __init__(self, **kargs):
        """Put self into game.robots and into the World."""
        fields = {
            "id": None,
            "location": None,
            "packages": []
        }
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 
        game.robots[self.id] = self
        self.square.robot = self

    def __getattr__(self, attr):
        """square is an attribute that returns the square we're on."""
        if attr == "square":
            return game.world.at(self.location)
        raise AttributeError

    def destruct(self):
        """Remove this robot from game.robots and the World.  Then del it.
        
        Unfortunately, renaming this class __del__ doesn't work.  It might be
        nice to delete the robot's packages, but it is not required for our
        current task.
        
        """
        self.square.robot = None
        del(game.robots[self.id])
        del(self)

    def move(self, direction):
        """Move the robot to an adjacent square.  
        
        This is not constrained.  This will update the World.
        
        """
        self.square.robot = None
        self.location = move(self.location, direction)
        self.square.robot = self

    def pick(self, package):
        """Add a package to our list of packages.
        
        If our Square knows about the Package, update the Square.  For 
        RobotOther's, the Square might not always know about the Package.
        
        """  
        self.packages.append(package)
        if package in self.square.packagesEncountered:
            self.square.packagesEncountered.remove(package)

    def drop(self, package):
        """Drop a package from our list of packages."""
        self.packages.remove(package)

class RobotSelf(RobotBase):

    """The robot that this game controls.
    
    The following methods are defined:

    __init__ -- Set game.robotSelf to self.

    __getattr__ -- Calculate self.packagesWeight dynamically.

    """

    def __init__(self, **kargs):
        """Set game.robotSelf to self."""
        fields = {
            "maximumCarryingCapacity": None,
            "money": None
        }
        fields.update(kargs)
        RobotBase.__init__(self, **fields) 
        game.robotSelf = self

    def __getattr__(self, attr):
        """Calculate self.packagesWeight dynamically."""
        if attr == "packagesWeight":
            return reduce(operator.add, [i.weight for i in self.packages], 0)
        return RobotBase.__getattr__(self, attr)

class RobotOther(RobotBase):

    """A robot that this game competes against."""

    def __init__(self, **kargs):
        fields = {
            # I expect there to be more stuff here later.
        }
        fields.update(kargs)
        RobotBase.__init__(self, **fields) 

class Package(AutomaticClass):

    """This is a simple class to represent a package.
    
    The following methods are defined:

    __init__ -- Put self into game.packages.
    
    """

    def __init__(self, **kargs):
        """Put self into game.packages.

        If you know the location of the package, you should add it to 
        game.world yourself.
        
        """
        fields = {
            "id": None,
            "weight": None,
            "destination": None
        }
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 
        game.packages[self.id] = self

class Game(AutomaticClass):

    """This class ties a lot of the other classes together."""

    def __init__(self, **kargs):
        fields = {
            "world": None,
            "robotSelf": None,
            "robots": {}, # This includes robotSelf.
            "packages": {}
        }
        fields.update(kargs)
        AutomaticClass.__init__(self, **fields) 

game = Game()

# This is for testing.
def _packagePrint(robotSelf):
    f = lambda packages: [i.id for i in packages]
    print "  robotSelf.packages = %s" % f(robotSelf.packages)
    print "  robotSelf.packagesWeight = %s" % robotSelf.packagesWeight
    print "  robotSelf.square.packagesEncountered = %s" % f(
        robotSelf.square.packagesEncountered)
    
# Do some testing.
if __name__ == '__main__':

    for i in DIRECTIONS:
        print "Move (2, 2) %s: %s" % (i, move((2,2), i))
    print "Places adjacent to (2, 2): %s" % getAdjacent((2, 2))

    game.world = World(width = 5, height = 2)
    print "World(width = 5, height = 2):"
    print game.world

    robotBase = RobotBase(id = 4, location = (2, 2))
    print "game.robots should have 1 robot: %s" % len(game.robots.keys())
    print "robotBase.square is a '%s'" % robotBase.square
    robotBase.move(WEST)
    print "robotBase.move(WEST) from (2, 2) -> %s," % `robotBase.location`
    print "and World now shows a robot there: '%s'," % game.world[2][1]
    print "but not in previous location: '%s'" % game.world[2][2]
    robotBase.destruct()
    print "Now, after deleting the robot, (2, 1) -> '%s'," % game.world[2][1]
    print "and game.robots should not reference it.  Does it?  %s" % (
        game.robots.has_key(4))

    robotSelf = RobotSelf(id = 5, location = (1, 1))
    package = Package(id = 1, weight = 1)
    robotSelf.square.packagesEncountered.append(package)
    print "Before picking:"; _packagePrint(robotSelf)
    robotSelf.pick(package)
    print "After picking:"; _packagePrint(robotSelf)
    robotSelf.drop(package)
    print "After dropping (I should not put the package back in the Square):"
    _packagePrint(robotSelf)
