#!/usr/bin/python2

"""This file contains the GeneticAlgorithm class.

See usage for more information.

The following classes are defined: 

GeneticAlgorithm -- This class will find an optimum set of traits.

The following functions are defined:

usage -- Print out usage information to the user and exit.

getInitialTraits -- Given a traitsCount, return an initialTraits.

constrain -- Constrain a float to [0, 1].

unit_float -- Convert a value to a float.  

"""

import sys
import os
import os.path
import random

def usage():
    """Print out usage information to the user and exit."""
    sys.stderr.write(
        ("usage: %s external_program inital_traits...\n" % progname) +
        ("       %s external_program traits_count\n" % progname))
    sys.exit(1)

def getInitialTraits(traitsCount):
    """Given a traitsCount, return an initialTraits.

    I'll use all 0.5's since 0.5 is in the middle of [0, 1].
    
    """
    return [0.5] * traitsCount

def constrain(x): 
    """Constrain a float to [0, 1]."""
    return min([1.0, max([0.0, x])])

def unit_float(value):
    """Convert a value to a float.  
    
    If the value is not in [0, 1], raise ValueError.

    """
    new_value = float(value)
    if new_value != constrain(new_value):
        raise ValueError
    return new_value

class GeneticAlgorithm:

    """This class will find an optimum set of traits.

    Given an external program that accepts a list of traits (which will be
    represented as float values [0, 1]) and which returns as its last line of
    input a float value ([0, 1] where 1 is the most desireable), find the list
    of traits that performs optimally.

    The following methods are available:

    __init__ -- Accept input parameters.

    run -- Try to find the best traits.  This function may loop forever.

    The following private methods are used:

    _getTraitsScore -- Return the score for a set of traits.

    _getMutations -- Return a list of mutations to try.

    _getRandomMutation -- Return a completely random mutation.

    _getSlightVarianceMutation -- Change one trait randomly, return the 
    mutation.

    _getInheritedMutation -- Get one mutation based on intelligent guessing.

    The following private variables are used:

    _externalProgram -- The path to the external program mentioned above.

    _currentTraits -- The current best set of values for each of the traits.

    _currentTraitsScore -- The score received by _currentTraits.

    _previousTraits -- This is the previous best set of values for each of the
    traits.
    
    _generator -- A generator from the random module.

    """

    def __init__(self, externalProgram, initialTraits = []):
        """Accept input parameters."""
        self._externalProgram = externalProgram
        self._currentTraits = initialTraits
        self._currentTraitsScore = self._getTraitsScore(self._currentTraits)
        self._previousTraits = self._currentTraits[:]
        self._generator = random.Random()

    def run(self):
        """Try to find the best traits.  This function may loop forever.
        
        Use self._currentTraits to generate a set of mutations.  Test to see
        if any of those mutations are better than self._currentTraits.  If 
        any of them are, choose the best, update self._currentTraits and
        self._previousTraits, then begin again.  If none of them are, update
        self._previousTraits (so that we don't use self._getInheritedMutation 
        again) and begin again.

        This function is one of the few imperative functions in this program.
        :-/

        """
        while self._currentTraitsScore < 1.0:
            foundBetter = 0
            mutations = self._getMutations()
            for i in mutations:
                score = self._getTraitsScore(i)
                if score > self._currentTraitsScore:
                    self._currentTraitsScore = score
                    self._previousTraits = self._currentTraits
                    self._currentTraits = i
                    foundBetter = 1
                    print "%s = %s" % (i, score)
            if not foundBetter:
                self._previousTraits = self._currentTraits

    def _getTraitsScore(self, traits):
        """Return the score for a set of traits."""
        cmd = self._externalProgram + " " + " ".join(map(str, traits))
        p = os.popen(cmd)
        try:
            value = p.readlines()[-1][:-1]
            score = unit_float(value)
        except ValueError:
            sys.stderr.write("%s: %s returned %s!\n" % 
                (progname, self._externalProgram, value))
            sys.exit(1)
        except:
            sys.stderr.write("%s: call to %s failed\n" % 
                (progname, self._externalProgram))
            sys.exit(1)
        p.close()
        return score

    def _getMutations(self):
        """Return a list of mutations to try.
        
        The list will contain range of mutations based on the different 
        algorithms.  We'll use 0.5 and 2.0 for ratios given to 
        self._getInheritedMutation in order to do a exponential search for the 
        optimal adjustment.
        
        """
        return filter(lambda x: x, # Filter out None's.
            [self._getRandomMutation(),
             self._getSlightVarianceMutation(),
             self._getInheritedMutation(0.5), 
             self._getInheritedMutation(2.0)])

    def _getRandomMutation(self):
        """Return a completely random mutation.

        I'll round each random number to 7 (7 is somewhat arbitrary) digits
        which will move the random number from [0, 1) to [0, 1]. 
        
        """
        return [round(self._generator.random(), 7)
                for i in self._currentTraits]

    def _getSlightVarianceMutation(self):
        """Change one trait randomly, return the mutation.

        I will alter one trait randomly.  Note, when I alter a trait, I'll
        round it to 7 (7 is somewhat arbitrary) digits which will move the
        random number from [0, 1) to [0, 1]. 
        
        """
        newTraits = self._currentTraits[:]
        toAlter = self._generator.choice(range(0, len(newTraits)))
        newValue = round(self._generator.random(), 7)
        newTraits[toAlter] = newValue 
        return newTraits

    def _getInheritedMutation(self, ratio):
        """Get one mutation based on intelligent guessing.

        If self._previousTraits = self._currentTraits, I will return None.
        Otherwise, each trait of the mutation will be defined as:
        
            newTrait = constrain(
                previousTrait + ratio * (currentTrait - previousTrait))

        """
        if self._previousTraits == self._currentTraits:
            return None
        return [
            constrain(self._previousTraits[i] + ratio * (
                 self._currentTraits[i] - self._previousTraits[i]))
            for i in range(0, len(self._currentTraits))
        ]

if __name__ == '__main__':
    progname = os.path.basename(sys.argv[0])
    argvlen = len(sys.argv)
    if argvlen < 3:
        usage()
    externalProgram = sys.argv[1]
    if argvlen == 3:
        try: 
            initialTraits = getInitialTraits(int(sys.argv[2]))
            if not len(initialTraits):
                raise ValueError
        except ValueError: 
            sys.stderr.write("%s: traits_count must be a positive int\n" % 
                progname)
            sys.exit(1)
    else:
        try:
            initialTraits = map(unit_float, sys.argv[2:])
        except:
            sys.stderr.write("%s: each trait must be a float [0, 1]\n" % 
                progname)
            sys.exit(1)
    GeneticAlgorithm(externalProgram, initialTraits).run()
