"""This file contains the strategy for the game.

The following classes are defined:

WeightedShortestPath -- This encapsulates Dijkstra's weighted shortest path
algorithm.

Strategy -- This is a simple strategy based on WeightedShortestPath.

The following constants are defined:

ROBOPHOBIA, AQUAPHOBIA, LAZINESS, DILIGENCE, OPTIMISM -- These are silly names
for values in the range [0, 1] that will be used to weight paths.  I'm 
hoping to have time to use GeneticAlgorithm.py to fine tune my algorithm.

"""

ROBOPHOBIA = 0.5 
AQUAPHOBIA = 1.0
LAZINESS = 0.1
DILIGENCE = 0.7
EXCESSIVE = 0.5
OPTIMISM = 0.4

import DataModel
from DataModel import game

class WeightedShortestPath:

    """This encapsulates Dijkstra's weighted shortest path algorithm.
    
    The following methods are defined:

    update -- Update each Square with a new shortestPath and weightedValue.

    updateSquare -- Set the shortest path to a destination.

    getLocationValue -- Return the value of a location.

    The following private methods are defined:

    _clear -- Clear each Square's shortestPath and weightedValue.

    """

    def update(self):
        """Update each Square with a new shortestPath and weightedValue."""
        self._clear()
        for y in range(1, game.world.actualHeight - 1):
            for x in range(1, game.world.actualWidth - 1):
                self.updateSquare((x, y))

    def updateSquare(self, destination, _curr = None, _path = [],   
        _value = 0.0):
        """Set the shortest path to a destination.
        
        _curr, _path, _value -- Please ignore these.

        Start from game.robotSelf.location.    

        """
        # Think of game.robotSelf.location -> _curr? -> destination.
        if not _curr:
            _curr = game.robotSelf.location
            square = game.world.at(_curr)
            square.shortestPath = []
            square.weightedValue = 0
        else:
            square = game.world.at(_curr)
            if square.terrain in [DataModel.WALL, DataModel.WATER]:
                square.shortestPath = None
                square.weightedValue = None
                return
            _value += self.getLocationValue(_curr)
            _path.append(_curr)
            if (square.weightedValue and
                square.weightedValue <= _value):
                return
            square.shortestPath = _path
            square.weightedValue = _value
            if _curr == destination:
                return
        for i in DataModel.getAdjacent(_curr):
            self.updateSquare(destination, i, _path[:], _value)

    def getLocationValue(self, location):
        """Return the value of a location.
        
        This value will be a float.  Greater values are less desireable.
        
        """
        value = 1.0 + LAZINESS
        for i in DataModel.getAdjacent(location):
            square = game.world.at(i)
            robot = square.robot
            if robot and robot.id != game.robotSelf.id:
                value += ROBOPHOBIA
            if square.terrain == DataModel.WATER:
                value += AQUAPHOBIA
            if i in [j.destination for j in game.robotSelf.packages]:
                value -= DILIGENCE
            weightRatio = (float(game.robotSelf.packagesWeight) / 
                game.robotSelf.maximumCarryingCapacity)
            if ((square.terrain == DataModel.HOME_BASE or 
                 len(square.packagesEncountered)) and
                weightRatio < OPTIMISM):
                value -= OPTIMISM
        return value

    def _clear(self):
        """Clear each Square's shortestPath and weightedValue."""
        for y in range(1, game.world.actualHeight - 1):
            for x in range(1, game.world.actualWidth - 1):
                square = game.world[y][x]
                square.shortestPath = None
                square.weightedValue = None

class Strategy:

    """This is a simple strategy based on WeightedShortestPath.
    
    The idea is simple, if you have any packages, start finding the optimal
    way to drop them off.  Otherwise, starting finding the optimal way to 
    pick some up.  This is really too simple, but it's a start, and I'm almost
    out of time!

    The following methods are defined:

    getMove -- Get the next move.

    The following private methods are defined:

    _getHomeBases -- Get a list of all the HOME_BASE's.
    
    """

    def getMove(self):
        """Get the next move."""
        WeightedShortestPath().update()
        if len(game.robotSelf.packages):
            destinations = [i.destination for i in game.robotSelf.packages]
        else:
            destinations = self._getHomeBases()
        bestDestination = None
        for i in destinations:
            square = game.world.at(i)
            if (bestDestination == None or 
                square.weightedValue < bestDestination.weightedValue):
                if (not square.robot or # Hackish.
                    square.robot.id != game.robotSelf.id):
                    bestDestination = square
        firstStep = bestDestination.shortestPath[0]
        return DataModel.inverseMove(game.robotSelf.location, firstStep)
            
    def _getHomeBases(self):
        """Get a list of all the HOME_BASE's."""
        return [(x, y)
            for y in range(1, game.world.actualHeight - 1)
                for x in range(1, game.world.actualWidth - 1)
                    if game.world[y][x].terrain == DataModel.HOME_BASE]
